%% @file
% combined.m
%
% @author Jose Vicente and Lars Johannesen
%
% @section DISCLAIMER
% This software and documentation (the "Software") were developed at the Food
% and Drug Administration (FDA) by employees of the Federal Government in the
% course of their official duties. Pursuant to Title 17, Section 105 of the
% United States Code, this work is not subject to copyright protection and is
% in the public domain. Permission is hereby granted, free of charge, to any
% person obtaining a copy of the Software, to deal in the Software without
% restriction, including without limitation the rights to use, copy, modify,
% merge, publish, distribute, sublicense, or sell copies of the Software or
% derivatives, and to permit persons to whom the Software is furnished to do
% so. FDA assumes no responsibility whatsoever for use by other parties of
% the Software, its source code, documentation or compiled executables, and
% makes no guarantees, expressed or implied, about its quality, reliability,
%  or any other characteristic. Further, use of this code in no way implies
% endorsement by the FDA or confers any advantage in regulatory decisions.
% Although this software can be redistributed and/or modified freely, we
% ask that any derivative works bear some notice that they are derived from
% it, and any modified versions bear some notice that they have been modified.
%

%%
% @brief Returns an array containing the locations of the beats detected using the different signals on the record
% @param rec Physionet record containing at least on ECG lead and other biologiocal signals
% @param debug Debug flag, if true prints error and warning messages
% @return [anns,out] anns: annotations with the location of detected beats in ms; out: record's signals information
%
function [varargout] = combined(rec, varargin)
%combined Returns an array containing the locations of the beats detected using the different signals on the record
%
% rec Physionet record containing at least on ECG lead and other biologiocal signals
%
%                                  DISCLAIMER                                  
% This software and documentation (the "Software") were developed at the Food  
% and Drug Administration (FDA) by employees of the Federal Government in the  

% United States Code, this work is not subject to copyright protection and is  
% in the public domain. Permission is hereby granted, free of charge, to any   
% person obtaining a copy of the Software, to deal in the Software without     
% restriction, including without limitation the rights to use, copy, modify,   
% merge, publish, distribute, sublicense, or sell copies of the Software or    
% derivatives, and to permit persons to whom the Software is furnished to do   
% so. FDA assumes no responsibility whatsoever for use by other parties of     
% the Software, its source code, documentation or compiled executables, and    
% makes no guarantees, expressed or implied, about its quality, reliability,   
%  or any other characteristic. Further, use of this code in no way implies    
% endorsement by the FDA or confers any advantage in regulatory decisions.     
% Although this software can be redistributed and/or modified freely, we       
% ask that any derivative works bear some notice that they are derived from    
% it, and any modified versions bear some notice that they have been modified. 
%                                                                              
% Written by: Jose Vicente and Lars Johannesen

	bp_transit = -140; % msec
	ppg_transit = -330; % msec
	sv_transit = -450; % msec
	window = 0; % msec, 0 = disabled (mark_peaks)
	prim_minsqi = 1;
	maxlag = 3000; % msec
	min_gap = 1500;%minimum gap length in msec
	border_thresh = 250; %closest distance to existing beat to start to fill in a gap
	wavf = gaussian(150, 0.04);  % Width in ms
	estlag = 1; % correct for transit 1/0 yes/no
	correcttransit = {'SV','PPG','BP'}; % Which signals to correct
	estratio = 0.8; % Lag peak inside window must be greater than peak outside window by x
	expectedavghr = 60; % Expected avg hr
	expectedmaxhr = 250; % Expected max hr
	minbeats = 0.3; % A detector must have at least have this amount of detections relative to total time and expectedavghr

	chanweights = struct('ECG',100,...
			     'BP',20,...
			     'SV',1.1,...
			     'PPG',2,...
			     'EOG',1,...
			     'EEG',1.5,...
			     'RR',0);

	opt = struct('bp_transit', bp_transit, ...
		     'sv_transit', sv_transit,...
		     'ppg_transit', ppg_transit,...
		     'window',window,...
		     'prim_minsqi',prim_minsqi,...
		     'maxlag',maxlag,...
		     'min_gap',min_gap,...
		     'estlag',estlag,...
		     'estratio',estratio,...
		     'border_thresh',border_thresh,...
		     'expectedavghr',expectedavghr,...
		     'expectedmaxhr',expectedmaxhr,...
		     'minbeats',minbeats);
	opt.correcttransit = correcttransit;
	opt.chanweights = chanweights;

	if nargout == 1 && nargin == 0
		varargout{1} = opt;
		return;
	end

	debug = 0;
	if nargin >= 2
		debug = varargin{1};
	end

	if nargin >= 3
		opt = varargin{2};
	end

	%Load dictionary of known signals
	[signalsdictionary.names,signalsdictionary.standardname,signalsdictionary.types]=textread('dictionary.csv','%s %s     %s','delimiter',',','headerlines',1);
	
	total_time = (length(rec{1}.data)/rec{1}.fs)*1000; % in ms

	%Process all signals
	for i=1:length(rec)
		success = false;
		out{i}.anns = [];
		out{i}.raw_anns = [];
		out{i}.raw_sqi = [];
		out{i}.sqi = [];
		out{i}.originalname = rec{i}.name; %Original name in the recording
		out{i}.fs = rec{i}.fs; %Signal's sampling frequency
		out{i}.type = '';
		out{i}.data = [];
		out{i}.lag = 0;
		out{i}.ratio = 0;
		out{i}.length = length(rec{i}.data);
		out{i}.name = '';

		try
			[lookupname, idx, success] = search_dictionary(upper(rec{i}.name),signalsdictionary.names);
		catch
			if(debug)
				disp('Dictionary failed');
			end
			
			if (i == 1)
				[lookupname, idx, success] = search_dictionary('ECG',signalsdictionary.names);
			end
		end

		if (success)
			out{i}.type = signalsdictionary.types{idx}; %Type of known signals
			out{i}.name = lookupname; %Corresponding known (standard) name from the dictionary
			out{i}.standardname = signalsdictionary.standardname{idx};

			try 
				%Process detections accordingly to the type of signal
				if ((i==1) || strcmp(out{i}.type,'ECG'))
					if rec{i}.fs ~= 250
						out{i}.data = resample(rec{i}.data, 250, rec{i}.fs);
					else
						out{i}.data = rec{i}.data;
					end

					out{i}.fs = 250;
					[anns,sqi] = ecg_detector(out{i}.data,0);

					if(~isempty(anns) && ~isempty(sqi))
						out{i}.anns = anns;

						%Convert annotations from samples to time
						out{i}.anns = out{i}.anns * 1000./out{i}.fs;
						out{i}.raw_anns = anns;
						out{i}.raw_sqi = sqi;
						if(opt.window > 0)
							out{i}.sqi = mark_peaks(out{i}.anns, sqi, opt.window);
						else
							out{i}.sqi = sqi;
						end
					end
				elseif (strcmp(out{i}.type,'BP'))
					if rec{i}.fs ~= 250
						out{i}.data = resample(rec{i}.data, 250, rec{i}.fs);
					else
						out{i}.data = rec{i}.data;
					end

					[anns,sqi] = abp_detector(out{i}.data);

					if(~isempty(anns) && ~isempty(sqi))
						out{i}.fs = 250;

						out{i}.raw_anns = anns;
						out{i}.raw_sqi = sqi;

						%Convert annotations from samples to time
						idx = anns>abs(opt.bp_transit);
						anns = anns(idx);
						sqi = sqi(idx);
						out{i}.anns = (anns * 1000./out{i}.fs) + opt.bp_transit;
						out{i}.lag = opt.bp_transit;
						if(opt.window > 0)
							out{i}.sqi = mark_peaks(out{i}.anns, sqi, opt.window);
						else
							out{i}.sqi = sqi;
						end
					end
				elseif (strcmp(out{i}.type,'PPG'))
					if rec{i}.fs ~= 250
						out{i}.data = resample(rec{i}.data, 250, rec{i}.fs);
					else
						out{i}.data = rec{i}.data;
					end

					[anns,sqi] = pleth_detector(out{i}.data);

					if(~isempty(anns) && ~isempty(sqi))
						out{i}.fs = 250;

						out{i}.raw_anns = anns;
						out{i}.raw_sqi = sqi;

						%Convert annotations from samples to time
						idx = anns>abs(opt.ppg_transit);
						anns = anns(idx);
						sqi = sqi(idx);
						out{i}.anns = (anns * 1000./out{i}.fs) + opt.ppg_transit;
						out{i}.lag = opt.ppg_transit;
						if(opt.window > 0)
							out{i}.sqi = mark_peaks(out{i}.anns, sqi, opt.window);
						else
							out{i}.sqi = sqi;
						end
					end
				elseif (strcmp(out{i}.type,'SV'))
					if rec{i}.fs ~= 250
						out{i}.data = resample(rec{i}.data, 250, rec{i}.fs);
					else
						out{i}.data = rec{i}.data;
					end

					[anns,sqi] = sv_detector(out{i}.data);
					if(~isempty(anns) && ~isempty(sqi))
						out{i}.fs = 250;

						out{i}.raw_anns = anns;
						out{i}.raw_sqi = sqi;

						%Convert annotations from samples to time
						idx = anns>abs(opt.sv_transit);
						anns = anns(idx);
						sqi = sqi(idx);
						out{i}.anns = (anns * 1000./out{i}.fs)+opt.sv_transit;
						out{i}.lag = opt.sv_transit;

						if opt.window > 0
							out{i}.sqi = mark_peaks(out{i}.anns, sqi, opt.window);
						else
							out{i}.sqi = sqi;
						end
					end
				elseif (strcmp(out{i}.type,'EEG'))
					if rec{i}.fs ~= 250
						out{i}.data = resample(rec{i}.data, 250, rec{i}.fs);
					else
						out{i}.data = rec{i}.data;
					end

					out{i}.fs = 250;
					out{i}.anns = [];
					out{i}.sqi = [];
				elseif (strcmp(out{i}.type,'EOG'))
					if rec{i}.fs ~= 250
						out{i}.data = resample(rec{i}.data, 250, rec{i}.fs);
					else
						out{i}.data = rec{i}.data;
					end

					out{i}.fs = 250;
					out{i}.anns = [];
					out{i}.sqi = [];
				end
			catch 
				if(debug)
					fprintf('Could not resample/process signal: %d (%s=>%s) - with detector: %s\n',i,out{i}.originalname,out{i}.name,out{i}.type);
				end
				out{i}.fs = rec{i}.fs; %Signal's sampling frequency
				out{i}.anns = []; 
			end
		end % End if success
	end % End foreach signal

	% Determine base annotations
        try
		% Take ECG record with most beats that satisfy:
		% >30% of the beats exepcted with 60 bpm < a constant HR of 250 bpm
		tsec = total_time / 1000 / 60; % time in min
		nbeats = tsec * opt.expectedavghr;
		minbeats = nbeats * opt.minbeats;
		maxbeats = tsec * opt.expectedmaxhr;
		curmax = 0;

		global_anns_init = [];
		for j=1:length(out)
			if strcmp(out{j}.type,'ECG')
				if ~isempty(out{j}.anns)
					curn = length(out{j}.anns);

					if curn > minbeats && curn < maxbeats && curn > curmax
						global_anns_init = out{j}.anns;
						curmax = curn;
					end
				end
			end
		end
        catch
		if(debug)
			disp('Could not determine base annotations');
		end
        end

	% Correct for transit time
	if opt.estlag == 1
		try 
			if ~isempty(global_anns_init)
				for i=2:length(out)
					if isfield(out{i},'anns') && ~isempty(out{i}.anns) 
						for j=1:length(opt.correcttransit)
							if strcmp(out{i}.type, opt.correcttransit{j})
								raw_anns = out{i}.raw_anns * 1000 / out{1}.fs;
								[lag, ratio, aux] = est_lag(global_anns_init, raw_anns, total_time, wavf, opt.maxlag, [-700 0]);
								out{i}.ratio = ratio;

								% SV/PPG etc has to be after ECG
								if lag < 0 && lag > -700 && ratio > opt.estratio
									nanns = out{i}.raw_anns;
									sqi = out{i}.raw_sqi;

									idx = nanns>abs(lag);
									nanns = nanns(idx);
									sqi = sqi(idx);
									out{i}.anns = (nanns * 1000./out{i}.fs)+lag;
									out{i}.lag = lag;
									if opt.window > 0
										out{i}.sqi = mark_peaks(nanns, sqi, opt.window);
									else
										out{i}.sqi = sqi;
									end
								end
							end
						end
					end
				end
			end
		catch
			if(debug)
				fprintf('Could not determine transit time\n');
			end
		end
	end

	% Strategy 1 - Use signal 1 + other signals with good quality for majority voting
	try 
		% Setup majority voting

		% Channel weighting
		chanweight = zeros(1,length(out));
		fnams = fieldnames(opt.chanweights);
		for j = 1:length(out)
			for jj=1:length(fnams)
				if strcmp(out{j}.type, fnams{jj})
					chanweight(j) = opt.chanweights.(fnams{jj});
				end
			end
		end

		% Options
		majoptions.mergemargin = 250;
		majoptions.absthres = 1.8;
		majoptions.relthres = 0.5;
		majoptions.winsize = 250;
		majoptions.windecay = 0.75;
		majoptions.len = total_time;
		majoptions.winwave = tukeywin(majoptions.winsize, majoptions.windecay);

		% Prepare annotations, SQI etc
		maj_anns = cellarrayunpack(out, 'anns');
		maj_sqi = cellarrayunpack(out, 'sqi');

		global_anns_step1a = majorityvote(maj_anns, majoptions, chanweight, maj_sqi);

		% Estimate RR series output - full, fill
		[~,rrs] = rrestimate(global_anns_step1a, [], total_time);

		% Setup majority voting with RR

		% Channel weighting
		rrweight = opt.chanweights;
		if(isfield(opt.chanweights,'RR'))
			rrweight = opt.chanweights.RR;
		end
		rr_chanweight = [chanweight rrweight];

		% Setup options
		rr_majoptions.winwave(1:length(out)) = {majoptions.winwave};
		rr_majoptions.winwave(length(out)+1) = {tukeywin(440,0.05)};

		% Prep input
		rr_anns = maj_anns;
		rr_anns{length(maj_anns)+1} = rrs;

		rr_sqi = maj_sqi;
		rr_sqi{length(maj_sqi)+1} = ones(1,length(rrs));

		global_anns_step1b = majorityvote(rr_anns, rr_majoptions, rr_chanweight, rr_sqi);
	catch
		if(debug)
			disp('Crashed when trying to fill gaps');
		end

		global_anns_step1a = global_anns_init;
		global_anns_step1b = global_anns_init;
	end

	% Strategy 2 - Try and use EEG/EOG
	try 
		% Detect in EEG/EOG signals
		for j=2:length(out)
			if (strcmp(out{j}.type,'EEG') && ~isempty(out{j}.data)) || (strcmp(out{j}.type,'EOG') && ~isempty(out{j}.data))
				tanns = eeg_detector(out{j}.data, global_anns_step1b);
				sqi = ones(1,length(tanns));

				if (~isempty(tanns))
					out{j}.anns = tanns * 1000. / out{j}.fs;
					out{j}.raw_anns = tanns;
					out{j}.raw_sqi = sqi;
					out{j}.sqi = sqi;

					[lag, ratio, aux] = est_lag(global_anns_step1b, out{j}.anns,total_time, wavf, opt.maxlag, [-75 75]);
					out{j}.lag = lag;
					out{j}.ratio = ratio;

					% If the delay is not as anticipated ignore it
					if abs(lag) > 75 || ratio < opt.estratio
						out{j}.anns = [];
						out{j}.sqi = [];
					end
				end
			end
		end

		% Prepare annotations, SQI etc
		maj_anns = cellarrayunpack(out, 'anns');
		maj_sqi = cellarrayunpack(out, 'sqi');

		% Prep input
		rr_anns = maj_anns;
		rr_anns{length(maj_anns)+1} = rrs;

		rr_sqi = maj_sqi;
		rr_sqi{length(maj_sqi)+1} = ones(1,length(rrs));

		global_anns_step2 = majorityvote(rr_anns, rr_majoptions, rr_chanweight, rr_sqi);
	catch
		if(debug)
			disp('Crashed when trying to fill gaps with eeg/eog');
		end

		global_anns_step2 = global_anns_step1b;
	end

	varargout{1} = global_anns_step1a;
	varargout{2} = out;
	varargout{3} = opt;
end
